# Copyright 2024 VAST Data Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import io
import unittest
from unittest import mock

import ddt
import requests

from manila import exception as manila_exception
from manila.share.drivers.vastdata import driver_util
from manila.share.drivers.vastdata import rest as vast_rest


fake_metrics = driver_util.Bunch.from_dict(
    {
        "object_ids": [1],
        "prop_list": [
            "timestamp",
            "object_id",
            "Capacity,drr",
            "Capacity,physical_space_in_use",
            "Capacity,physical_space",
            "Capacity,logical_space",
            "Capacity,logical_space_in_use",
        ],
        "data": [
            [
                "2024-04-13T14:47:48Z",
                1,
                1.2,
                30635076602.0,
                711246584217.0,
                505850953728.0,
                22370924820.0,
            ],
            [
                "2024-04-13T14:47:38Z",
                1,
                1.2,
                30635109399.0,
                711246584217.0,
                505850134528.0,
                22370810131.0,
            ],
            [
                "2024-04-13T14:47:28Z",
                1,
                1.2,
                30635142195.0,
                711246584217.0,
                505849217024.0,
                22370720020.0,
            ],
            [
                "2024-04-13T14:47:18Z",
                1,
                1.2,
                30635174991.0,
                711246584217.0,
                505848365056.0,
                22370654484.0,
            ],
            [
                "2024-04-13T14:47:08Z",
                1,
                1.2,
                30635207787.0,
                711246584217.0,
                505847447552.0,
                22420396308.0,
            ],
            [
                "2024-04-13T14:46:58Z",
                1,
                1.2,
                30635248783.0,
                711246584217.0,
                505846398976.0,
                22420306196.0,
            ],
        ],
        "granularity": None,
    }
)


class TestSession(unittest.TestCase):

    def setUp(self):
        self.session = vast_rest.Session(
            "host", "username",
            "password", False, "1.0"
        )

    @mock.patch("requests.Session.request")
    def test_refresh_auth_token_success(self, mock_request):
        mock_request.return_value.json.return_value = {"access": "test_token"}
        self.session.refresh_auth_token()
        self.assertEqual(
            self.session.headers["authorization"],
            "Bearer test_token"
        )

    @mock.patch("requests.Session.request")
    def test_refresh_auth_token_failure(self, mock_request):
        mock_request.side_effect = ConnectionError()
        with self.assertRaises(manila_exception.VastApiException):
            self.session.refresh_auth_token()

    @mock.patch("requests.Session.request")
    def test_request_success(self, mock_request):
        mock_request.return_value.status_code = 200
        self.session.request(
            "GET", "test_method",
            log_result=False, params={"foo": "bar"}
        )
        mock_request.assert_called_once_with(
            "GET", "https://host/api/test_method/",
            verify=False, params={"foo": "bar"}
        )

    @mock.patch("requests.Session.request")
    def test_request_failure_400(self, mock_request):
        mock_request.return_value.status_code = 400
        mock_request.return_value.text = "foo/bar"
        with self.assertRaises(manila_exception.VastApiException):
            self.session.request(
                "POST", "test_method",
                data={"data": {"foo": "bar"}}
            )

    def test_request_failure_500(self):

        resp = requests.Response()
        resp.status_code = 500
        resp.raw = io.BytesIO(b"Server error")

        with mock.patch(
                "requests.Session.request", new=lambda *a, **k: resp
        ):
            with self.assertRaises(manila_exception.VastApiException) as exc:
                self.session.request("GET", "test_method", log_result=False)
            self.assertIn("Server Error", str(exc.exception))

    def test_request_no_return_content(self):
        resp = requests.Response()
        resp.status_code = 200
        resp.raw = io.BytesIO(b"")

        with mock.patch(
                "requests.Session.request", new=lambda *a, **k: resp
        ):
            res = self.session.request("GET", "test_method")
        self.assertFalse(res)

    @mock.patch(
        "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
        mock.MagicMock()
    )
    def test_refresh_token_retries(self):
        resp = requests.Response()
        resp.status_code = 403
        resp.raw = io.BytesIO(b"Token is invalid")

        with mock.patch("requests.Session.request", new=lambda *a, **k: resp):
            with self.assertRaises(manila_exception.VastApiRetry):
                self.session.request("POST", "test_method", foo="bar")

    def test_getattr_with_underscore(self):
        with self.assertRaises(AttributeError):
            self.session.__getattr__("_private")

    @mock.patch.object(vast_rest.Session, "request")
    def test_getattr_without_underscore(self, mock_request):
        attr = "public"
        params = {"key": "value"}
        self.session.__getattr__(attr)(**params)
        mock_request.assert_called_once_with("get", attr, params=params)


class TestVastResource(unittest.TestCase):
    def setUp(self):
        self.mock_rest = mock.MagicMock()
        self.vast_resource = vast_rest.VastResource(self.mock_rest)

    def test_list_with_filtering_params(self):
        self.vast_resource.list(name="test")
        self.mock_rest.session.get.assert_called_with(
            self.vast_resource.resource_name, params={"name": "test"}
        )

    def test_create_with_provided_params(self):
        self.vast_resource.create(name="test", size=10)
        self.mock_rest.session.post.assert_called_with(
            self.vast_resource.resource_name, data={"name": "test", "size": 10}
        )

    def test_update_with_provided_params(self):
        self.vast_resource.update("1", name="test", size=10)
        self.mock_rest.session.patch.assert_called_with(
            f"{self.vast_resource.resource_name}/1",
            data={"name": "test", "size": 10}
        )

    def test_delete_when_entry_not_found(self):
        self.vast_resource.one = mock.MagicMock(return_value=None)
        self.vast_resource.delete("test")
        self.mock_rest.session.delete.assert_not_called()

    def test_delete_when_entry_found(self):
        mock_entry = mock.MagicMock()
        mock_entry.id = "1"
        self.vast_resource.one = mock.MagicMock(return_value=mock_entry)
        self.vast_resource.delete("test")
        self.mock_rest.session.delete.assert_called_with(
            f"{self.vast_resource.resource_name}/{mock_entry.id}"
        )

    def test_one_when_no_entries_found(self):
        self.vast_resource.list = mock.MagicMock(return_value=[])
        result = self.vast_resource.one("test")
        self.assertIsNone(result)

    def test_one_when_multiple_entries_found(self):
        self.vast_resource.list = mock.MagicMock(
            return_value=[mock.MagicMock(), mock.MagicMock()]
        )
        with self.assertRaises(manila_exception.VastDriverException):
            self.vast_resource.one("test")

    def test_one_when_single_entry_found(self):
        mock_entry = mock.MagicMock()
        self.vast_resource.list = mock.MagicMock(return_value=[mock_entry])
        result = self.vast_resource.one("test")
        self.assertEqual(result, mock_entry)

    def test_ensure_when_entry_not_found(self):
        self.vast_resource.one = mock.MagicMock(return_value=None)
        mock_entry = mock.MagicMock()
        self.vast_resource.create = mock.MagicMock(return_value=mock_entry)
        result = self.vast_resource.ensure("test", size=10)
        self.assertEqual(result, mock_entry)

    def test_ensure_when_entry_found(self):
        mock_entry = mock.MagicMock()
        self.vast_resource.one = mock.MagicMock(return_value=mock_entry)
        result = self.vast_resource.ensure("test", size=10)
        self.assertEqual(result, mock_entry)


class ViewTest(unittest.TestCase):
    @mock.patch(
        "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
        mock.MagicMock()
    )
    def test_view_create(self):
        with mock.patch(
                "manila.share.drivers.vastdata.rest.Session.post"
        ) as mock_session:
            rest_api = vast_rest.RestApi(
                "host",
                "username",
                "password",
                True,
                "1.0"
            )
            rest_api.views.create("test-view", "/test", 1)

        self.assertEqual(("views",), mock_session.call_args.args)
        self.assertDictEqual(
            {
                "data": {
                    "name": "test-view",
                    "path": "/test",
                    "policy_id": 1,
                    "create_dir": True,
                    "protocols": ["NFS"],
                }
            },
            mock_session.call_args.kwargs,
        )


@mock.patch(
    "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
    mock.MagicMock()
)
@mock.patch(
    "manila.share.drivers.vastdata.rest.Session.get",
    mock.MagicMock(return_value=fake_metrics),
)
class TestCapacityMetrics(unittest.TestCase):

    def test_capacity_metrics(self):
        metrics_list = [
            "Capacity,drr",
            "Capacity,logical_space",
            "Capacity,logical_space_in_use",
            "Capacity,physical_space",
            "Capacity,physical_space_in_use",
        ]
        expected = {
            "": 1,
            "drr": 1.2,
            "physical_space_in_use": 30635248783.0,
            "physical_space": 711246584217.0,
            "logical_space": 505846398976.0,
            "logical_space_in_use": 22420306196.0,
        }
        rest_api = vast_rest.RestApi(
            "host",
            "username",
            "password",
            True,
            "1.0"
        )
        metrics = rest_api.capacity_metrics.get(metrics_list)
        self.assertDictEqual(expected, metrics)


@mock.patch(
    "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
    mock.MagicMock()
)
@ddt.ddt
class TestFolders(unittest.TestCase):

    @mock.patch(
        "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
        mock.MagicMock()
    )
    def setUp(self):
        self.rest_api = vast_rest.RestApi(
            "host", "username", "password", True, "1.0"
        )

    @ddt.data(
        "4.3.9",
        "4.0.11.12",
        "3.4.6.123.1",
        "4.5.6-1",
        "4.6.0",
        "4.6.0-1",
        "4.6.0-1.1",
        "4.6.9",
    )
    def test_requisite_decorator(self, cluster_version):
        """Test `requisite` decorator produces exception

         when cluster version doesn't met requirements
         """
        with mock.patch(
            "manila.share.drivers.vastdata.rest.RestApi.get_sw_version",
            new=lambda s: cluster_version,
        ):
            self.assertRaises(
                manila_exception.VastDriverException,
                lambda: self.rest_api.folders.delete("/abc")
            )

    def test_trash_api_disabled(self):
        def raise_http_err(*args, **kwargs):
            resp = requests.Response()
            resp.status_code = 400
            resp.raw = io.BytesIO(b"trash folder disabled")
            raise manila_exception.VastApiException(message=resp.text)

        with (
            mock.patch(
                "manila.share.drivers.vastdata.rest.Session.delete",
                side_effect=raise_http_err,
            ),
            mock.patch(
                "manila.share.drivers.vastdata.rest.RestApi.get_sw_version",
                new=lambda s: "5.0.0",
            ),
        ):
            with self.assertRaises(
                    manila_exception.VastDriverException
            ) as exc:
                self.rest_api.folders.delete("/abc")
        self.assertIn("Trash Folder Access is disabled", str(exc.exception))

    def test_trash_api_unpredictable_error(self):
        def raise_http_err(*args, **kwargs):
            raise RuntimeError()

        with (
            mock.patch(
                "manila.share.drivers.vastdata.rest.Session.delete",
                side_effect=raise_http_err,
            ),
            mock.patch(
                "manila.share.drivers.vastdata.rest.RestApi.get_sw_version",
                new=lambda s: "5.0.0",
            ),
        ):
            with self.assertRaises(RuntimeError):
                self.rest_api.folders.delete("/abc")

    def test_double_deletion(self):
        def raise_http_err(*args, **kwargs):
            resp = requests.Response()
            resp.status_code = 400
            resp.raw = io.BytesIO(b"no such directory")
            raise manila_exception.VastApiException(message=resp.text)

        with (
            mock.patch(
                "manila.share.drivers.vastdata.rest.Session.delete",
                side_effect=raise_http_err,
            ),
            mock.patch(
                "manila.share.drivers.vastdata.rest.RestApi.get_sw_version",
                new=lambda s: "5.0.0",
            ),
        ):
            with self.assertLogs(level="DEBUG") as cm:
                self.rest_api.folders.delete("/abc")
            self.assertIn(
                "remote directory might have been removed earlier",
                str(cm.output)
            )


class VipPoolTest(unittest.TestCase):
    @mock.patch(
        "manila.share.drivers.vastdata.rest.Session.refresh_auth_token",
        mock.MagicMock()
    )
    def setUp(self):
        self.rest_api = vast_rest.RestApi(
            "host",
            "username",
            "password",
            True,
            "1.0"
        )

    def test_no_vipool(self):
        with mock.patch(
                "manila.share.drivers.vastdata.rest.Session.get",
                return_value=[]
        ):
            with self.assertRaises(
                    manila_exception.VastDriverException
            ) as exc:
                self.rest_api.vip_pools.vips("test-vip")
        self.assertIn("No vip pool found", str(exc.exception))

    def test_no_vips(self):
        vippool = driver_util.Bunch(ip_ranges=[])
        with mock.patch(
            "manila.share.drivers.vastdata.rest.Session.get",
                return_value=[vippool]
        ):
            with self.assertRaises(
                    manila_exception.VastDriverException
            ) as exc:
                self.rest_api.vip_pools.vips("test-vip")
        self.assertIn(
            "Pool test-vip has no available vips",
            str(exc.exception)
        )

    def test_vips_ok(self):
        vippool = driver_util.Bunch(
            ip_ranges=[["15.0.0.1", "15.0.0.4"], ["10.0.0.27", "10.0.0.30"]]
        )
        expected = [
            "15.0.0.1",
            "15.0.0.2",
            "15.0.0.3",
            "15.0.0.4",
            "10.0.0.27",
            "10.0.0.28",
            "10.0.0.29",
            "10.0.0.30",
        ]
        with mock.patch(
            "manila.share.drivers.vastdata.rest.Session.get",
                return_value=[vippool]
        ):
            vips = self.rest_api.vip_pools.vips("test-vip")
        self.assertListEqual(vips, expected)


class TestRestApi(unittest.TestCase):

    @mock.patch("manila.share.drivers.vastdata.rest.Session")
    def test_get_sw_version(self, mock_session):
        mock_session.return_value.versions.return_value = [
            mock.MagicMock(sys_version="1.0")
        ]
        rest_api = vast_rest.RestApi(
            "host", "username", "password", True, "1.0"
        )
        version = rest_api.get_sw_version()
        self.assertEqual(version, "1.0")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from manila.policies import base


BASE_POLICY_NAME = 'resource_lock:%s'

DEPRECATED_REASON = """
The resource lock API now supports scope and default roles.
"""

deprecated_lock_get = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_lock_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_lock_get_all_projects = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all_projects',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_lock_create = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat'
)
deprecated_lock_update = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update',
    check_str=base.RULE_ADMIN_OR_OWNER_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_lock_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete',
    check_str=base.RULE_ADMIN_OR_OWNER_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)
deprecated_bypass_locked_show_action = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'bypass_locked_show_action',
    check_str=base.RULE_ADMIN_OR_OWNER_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2023.2/Bobcat',
)

# We anticipate bypassing is desirable only for resource visibility locks.
# Without a bypass, the lock would have to be set aside each time the lock
# owner wants to view the resource.

lock_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get',
        check_str=base.ADMIN_OR_SERVICE_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get details of a given resource lock.",
        operations=[
            {
                'method': 'GET',
                'path': '/resource-locks/{lock_id}'
            }
        ],
        deprecated_rule=deprecated_lock_get,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all',
        check_str=base.ADMIN_OR_SERVICE_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get all resource locks.",
        operations=[
            {
                'method': 'GET',
                'path': '/resource-locks'
            },
            {
                'method': 'GET',
                'path': '/resource-locks?{query}'
            }
        ],
        deprecated_rule=deprecated_lock_get_all,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all_projects',
        check_str=base.ADMIN_OR_SERVICE,
        scope_types=['project'],
        description="Get resource locks from all project namespaces.",
        operations=[
            {
                'method': 'GET',
                'path': '/resource-locks?all_projects=1'
            },
            {
                'method': 'GET',
                'path': '/resource-locks?all_projects=1&'
                        'project_id={project_id}'
            }
        ],
        deprecated_rule=deprecated_lock_get_all_projects,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create',
        check_str=base.ADMIN_OR_SERVICE_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Create a resource lock.",
        operations=[
            {
                'method': 'POST',
                'path': '/resource-locks'
            }
        ],
        deprecated_rule=deprecated_lock_create,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update',
        check_str=base.ADMIN_OR_SERVICE_OR_OWNER_USER,
        scope_types=['project'],
        description="Update a resource lock.",
        operations=[
            {
                'method': 'PUT',
                'path': '/resource-locks/{lock_id}'
            }
        ],
        deprecated_rule=deprecated_lock_update,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete',
        check_str=base.ADMIN_OR_SERVICE_OR_OWNER_USER,
        scope_types=['project'],
        description="Delete a resource lock.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/resource-locks/{lock_id}'
            }
        ],
        deprecated_rule=deprecated_lock_delete,
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'bypass_locked_show_action',
        check_str=base.ADMIN_OR_SERVICE_OR_OWNER_USER,
        scope_types=['project'],
        description="Bypass a visibility lock placed in a resource.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-access-rules/{share_access_id}'
            },
            {
                'method': 'GET',
                'path': ('/share-access-rules?share_id={share_id}'
                         '&key1=value1&key2=value2')
            },
        ],
        deprecated_rule=deprecated_bypass_locked_show_action,
    ),
]


def list_rules():
    return lock_policies



# Copyright 2012 IBM Corp.
# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import datetime
from unittest import mock
import webob

import ddt
from oslo_utils import timeutils

from manila.api.openstack import api_version_request as api_version
from manila.api.v2 import services
from manila import context
from manila import db
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes


fake_services_list = [
    {
        'binary': 'manila-scheduler',
        'host': 'host1',
        'availability_zone': {'name': 'manila1'},
        'id': 1,
        'disabled': True,
        'disabled_reason': 'test1',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 2),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 27),
    },
    {
        'binary': 'manila-share',
        'host': 'host1',
        'availability_zone': {'name': 'manila1'},
        'id': 2,
        'disabled': True,
        'disabled_reason': 'test2',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 5),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 27)},
    {
        'binary': 'manila-scheduler',
        'host': 'host2',
        'availability_zone': {'name': 'manila2'},
        'id': 3,
        'disabled': False,
        'disabled_reason': '',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 19, 6, 55, 34),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28)},
    {
        'binary': 'manila-share',
        'host': 'host2',
        'availability_zone': {'name': 'manila2'},
        'id': 4,
        'disabled': True,
        'disabled_reason': 'test4',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
    },
]


fake_response_service_list = {'services': [
    {
        'id': 1,
        'binary': 'manila-scheduler',
        'host': 'host1',
        'zone': 'manila1',
        'status': 'disabled',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 2),
    },
    {
        'id': 2,
        'binary': 'manila-share',
        'host': 'host1',
        'zone': 'manila1',
        'status': 'disabled',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 5),
    },
    {
        'id': 3,
        'binary': 'manila-scheduler',
        'host': 'host2',
        'zone': 'manila2',
        'status': 'enabled',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 19, 6, 55, 34),
    },
    {
        'id': 4,
        'binary': 'manila-share',
        'host': 'host2',
        'zone': 'manila2',
        'status': 'disabled',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
    },
]}

fake_response_service_list_with_disabled_reason = {'services': [
    {
        'id': 1,
        'binary': 'manila-scheduler',
        'host': 'host1',
        'zone': 'manila1',
        'status': 'disabled',
        'disabled_reason': 'test1',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 2),
    },
    {
        'id': 2,
        'binary': 'manila-share',
        'host': 'host1',
        'zone': 'manila1',
        'status': 'disabled',
        'disabled_reason': 'test2',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 5),
    },
    {
        'id': 3,
        'binary': 'manila-scheduler',
        'host': 'host2',
        'zone': 'manila2',
        'status': 'enabled',
        'disabled_reason': '',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 19, 6, 55, 34),
    },
    {
        'id': 4,
        'binary': 'manila-share',
        'host': 'host2',
        'zone': 'manila2',
        'status': 'disabled',
        'disabled_reason': 'test4',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
    },
]}
ENSURE_SHARES_VERSION = "2.86"


def fake_service_get_all(context):
    return fake_services_list


def fake_service_get_by_host_binary(context, host, binary):
    for service in fake_services_list:
        if service['host'] == host and service['binary'] == binary:
            return service
    return None


def fake_service_get_by_id(value):
    for service in fake_services_list:
        if service['id'] == value:
            return service
    return None


def fake_service_update(context, service_id, values):
    service = fake_service_get_by_id(service_id)
    if service is None:
        raise exception.ServiceNotFound(service_id=service_id)
    else:
        {'host': 'host1', 'binary': 'manila-share',
         'disabled': values['disabled']}


def fake_utcnow():
    return datetime.datetime(2012, 10, 29, 13, 42, 11)


@ddt.ddt
class ServicesTest(test.TestCase):

    def setUp(self):
        super(ServicesTest, self).setUp()

        self.mock_object(db, "service_get_all", fake_service_get_all)
        self.mock_object(timeutils, "utcnow", fake_utcnow)
        self.mock_object(db, "service_get_by_args",
                         fake_service_get_by_host_binary)
        self.mock_object(db, "service_update", fake_service_update)
        self.context = context.get_admin_context()
        self.controller = services.ServiceController()
        self.controller_legacy = services.ServiceControllerLegacy()
        self.resource_name = self.controller.resource_name
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))

    @ddt.data(
        ('os-services', '1.0', services.ServiceControllerLegacy),
        ('os-services', '2.6', services.ServiceControllerLegacy),
        ('services', '2.7', services.ServiceController),
        ('services', '2.83', services.ServiceController),
    )
    @ddt.unpack
    def test_services_list(self, url, version, controller):
        req = fakes.HTTPRequest.blank('/%s' % url, version=version)
        req.environ['manila.context'] = self.context

        res_dict = controller().index(req)

        if (api_version.APIVersionRequest(version) >=
                api_version.APIVersionRequest('2.83')):
            self.assertEqual(fake_response_service_list_with_disabled_reason,
                             res_dict)
        else:
            self.assertEqual(fake_response_service_list, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_host(self):
        req = fakes.HTTPRequest.blank('/services?host=host1', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][0],
            fake_response_service_list['services'][1],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_binary(self):
        req = fakes.HTTPRequest.blank(
            '/services?binary=manila-share', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][1],
            fake_response_service_list['services'][3],
        ]}

        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_zone(self):
        req = fakes.HTTPRequest.blank('/services?zone=manila1', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][0],
            fake_response_service_list['services'][1],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_status(self):
        req = fakes.HTTPRequest.blank(
            '/services?status=enabled', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][2],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_state(self):
        req = fakes.HTTPRequest.blank('/services?state=up', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][0],
            fake_response_service_list['services'][1],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_host_binary(self):
        req = fakes.HTTPRequest.blank(
            "/services?binary=manila-share&state=up", version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [fake_response_service_list['services'][1], ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    @ddt.data(
        ('os-services', '1.0', services.ServiceControllerLegacy),
        ('os-services', '2.6', services.ServiceControllerLegacy),
        ('services', '2.7', services.ServiceController),
    )
    @ddt.unpack
    def test_services_enable(self, url, version, controller):
        body = {'host': 'host1', 'binary': 'manila-share'}
        req = fakes.HTTPRequest.blank('/fooproject/%s' % url, version=version)

        res_dict = controller().update(req, "enable", body)

        self.assertEqual('enabled', res_dict['status'])
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    @ddt.data(
        ('os-services', '1.0', services.ServiceControllerLegacy),
        ('os-services', '2.6', services.ServiceControllerLegacy),
        ('services', '2.7', services.ServiceController),
        ('services', '2.83', services.ServiceController),
    )
    @ddt.unpack
    def test_services_disable(self, url, version, controller):
        req = fakes.HTTPRequest.blank(
            '/fooproject/%s/disable' % url, version=version)
        body = {'host': 'host1', 'binary': 'manila-share'}
        if (api_version.APIVersionRequest(version) >
                api_version.APIVersionRequest("2.83")):
            body['disabled_reason'] = 'test1'

        res_dict = controller().update(req, "disable", body)

        self.assertEqual('disabled', res_dict['status'])
        if (api_version.APIVersionRequest(version) >
                api_version.APIVersionRequest("2.83")):
            self.assertEqual(res_dict['disabled_reason'], 'test1')

        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    @ddt.data(
        ('os-services', '2.7', services.ServiceControllerLegacy),
        ('services', '2.6', services.ServiceController),
        ('services', '1.0', services.ServiceController),
    )
    @ddt.unpack
    def test_services_update_legacy_url_2_dot_7_api_not_found(self, url,
                                                              version,
                                                              controller):
        req = fakes.HTTPRequest.blank(
            '/fooproject/%s/fake' % url, version=version)
        body = {'host': 'host1', 'binary': 'manila-share'}

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            controller().update,
            req, "disable", body,
        )

    @ddt.data(
        ('os-services', '2.7', services.ServiceControllerLegacy),
        ('services', '2.6', services.ServiceController),
        ('services', '1.0', services.ServiceController),
    )
    @ddt.unpack
    def test_services_list_api_not_found(self, url, version, controller):
        req = fakes.HTTPRequest.blank('/fooproject/%s' % url, version=version)

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod, controller().index, req)

    def test_ensure_shares_no_host_param(self):
        req = fakes.HTTPRequest.blank(
            '/fooproject/services/ensure', version=ENSURE_SHARES_VERSION)
        body = {}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.ensure_shares,
            req,
            body
        )

    def test_ensure_shares_host_not_found(self):
        req = fakes.HTTPRequest.blank(
            '/fooproject/services/ensure', version=ENSURE_SHARES_VERSION)
        req_context = req.environ['manila.context']
        body = {'host': 'host1'}

        mock_service_get = self.mock_object(
            db, 'service_get_by_args',
            mock.Mock(side_effect=exception.NotFound())
        )

        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller.ensure_shares,
            req,
            body
        )
        mock_service_get.assert_called_once_with(
            req_context,
            body['host'],
            'manila-share'
        )

    def test_ensure_shares_conflict(self):
        req = fakes.HTTPRequest.blank(
            '/fooproject/services/ensure', version=ENSURE_SHARES_VERSION)
        req_context = req.environ['manila.context']
        body = {'host': 'host1'}
        fake_service = {'id': 'fake_service_id'}

        mock_service_get = self.mock_object(
            db,
            'service_get_by_args',
            mock.Mock(return_value=fake_service)
        )
        mock_ensure = self.mock_object(
            self.controller.service_api,
            'ensure_shares',
            mock.Mock(side_effect=webob.exc.HTTPConflict)
        )

        self.assertRaises(
            webob.exc.HTTPConflict,
            self.controller.ensure_shares,
            req,
            body
        )
        mock_service_get.assert_called_once_with(
            req_context,
            body['host'],
            'manila-share'
        )
        mock_ensure.assert_called_once_with(
            req_context, fake_service, body['host']
        )

    def test_ensure_shares(self):
        req = fakes.HTTPRequest.blank(
            '/fooproject/services/ensure', version=ENSURE_SHARES_VERSION)
        req_context = req.environ['manila.context']
        body = {'host': 'host1'}
        fake_service = {'id': 'fake_service_id'}

        mock_service_get = self.mock_object(
            db,
            'service_get_by_args',
            mock.Mock(return_value=fake_service)
        )
        mock_ensure = self.mock_object(
            self.controller.service_api, 'ensure_shares',
        )

        response = self.controller.ensure_shares(req, body)

        self.assertEqual(202, response.status_int)
        mock_service_get.assert_called_once_with(
            req_context,
            body['host'],
            'manila-share'
        )
        mock_ensure.assert_called_once_with(
            req_context, fake_service, body['host']
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from manila.api.validation import helpers
from manila.api.validation import parameter_types
from manila.api.validation import response_types
from manila.common import constants

CONF = cfg.CONF

# TODO(stephenfin): Reject additional properties in a future microversion
create_request_body = {
    'type': 'object',
    'properties': {
        'resource_lock': {
            'type': 'object',
            'properties': {
                'resource_id': {
                    'type': 'string',
                    'format': 'uuid',
                    'description': helpers.description(
                        'resource_lock_resource_id'
                    ),
                },
                'lock_reason': {
                    'type': ['string', 'null'],
                    'maxLength': 1023,
                    'description': helpers.description(
                        'resource_lock_lock_reason_optional'
                    ),
                },
                'resource_type': {
                    'type': ['string', 'null'],
                    'enum': constants.RESOURCE_LOCK_RESOURCE_TYPES + (None,),
                    'default': constants.SHARE_RESOURCE_TYPE,
                    'description': helpers.description(
                        'resource_lock_resource_type'
                    ),
                },
                'resource_action': {
                    'type': ['string', 'null'],
                    'enum': constants.RESOURCE_LOCK_RESOURCE_ACTIONS + (None,),
                    'default': constants.RESOURCE_ACTION_DELETE,
                    'description': helpers.description(
                        'resource_lock_resource_action_create_optional'
                    ),
                },
            },
            'required': ['resource_id'],
            'additionalProperties': False,
            'description': helpers.description('resource_lock_object'),
        },
    },
    'required': ['resource_lock'],
    'additionalProperties': True,
}

update_request_body = {
    'type': 'object',
    'properties': {
        'resource_lock': {
            'type': 'object',
            'properties': {
                'resource_action': {
                    'type': ['string', 'null'],
                    'enum': constants.RESOURCE_LOCK_RESOURCE_ACTIONS + (None,),
                    'description': helpers.description(
                        'resource_lock_resource_action_optional'
                    ),
                },
                'lock_reason': {
                    'type': ['string', 'null'],
                    'maxLength': 1023,
                    'description': helpers.description(
                        'resource_lock_lock_reason_optional'
                    ),
                },
            },
            'additionalProperties': False,
            'description': helpers.description('resource_lock_object'),
        },
    },
    'required': ['resource_lock'],
    'additionalProperties': True,
}

index_request_query = {
    'type': 'object',
    'properties': {
        'limit': parameter_types.multi_params({
            **parameter_types.non_negative_integer,
            'description': helpers.description('limit'),
        }),
        # NOTE(stephenfin): This is parsed by 'common.get_pagination_params'
        # but we ignore it. We may wish to uncomment this when that is no
        # longer the case
        # 'marker': parameter_types.multi_params({
        #     'type': ['string'],
        # }),
        'offset': parameter_types.multi_params({
            **parameter_types.non_negative_integer,
            'description': helpers.description('offset'),
        }),
        'sort_key': parameter_types.multi_params({
            'type': 'string',
            'default': 'created_at',
            'description': helpers.description('sort_key_resource_locks'),
        }),
        'sort_dir': parameter_types.multi_params({
            'type': 'string',
            'default': 'desc',
            # TODO(stephenfin): This should be an enum, but we currently treat
            # anything != 'desc' as 'asc'. We should make this stricter in a
            # future microversion.
            # 'enum': ['asc', 'desc'],
            'description': helpers.description('sort_dir'),
        }),
        'with_count': parameter_types.multi_params(parameter_types.boolean),
        'created_since': parameter_types.multi_params({
            'type': 'string',
            'format': 'date-time',
            'description': helpers.description('created_since_query'),
        }),
        'created_before': parameter_types.multi_params({
            'type': 'string',
            'format': 'date-time',
            'description': helpers.description('created_before_query'),
        }),
        'project_id': parameter_types.multi_params({
            'type': ['string', 'null'],
            'format': 'uuid',
            'description': helpers.description(
                'resource_lock_project_id_query'
            ),
        }),
        'user_id': parameter_types.multi_params({
            'type': ['string', 'null'],
            'format': 'uuid',
            'description': helpers.description('resource_lock_user_id_query'),
        }),
        'resource_id': parameter_types.multi_params({
            'type': ['string', 'null'],
            'format': 'uuid',
            'description': helpers.description(
                'resource_lock_resource_id_query'
            ),
        }),
        'resource_action': parameter_types.multi_params({
            'type': ['string', 'null'],
            'enum': constants.RESOURCE_LOCK_RESOURCE_ACTIONS + (None,),
            'description': helpers.description(
                'resource_lock_resource_action_query'
            ),
        }),
        'resource_type': parameter_types.multi_params({
            'type': ['string', 'null'],
            'enum': constants.RESOURCE_LOCK_RESOURCE_TYPES + (None,),
            'description': helpers.description(
                'resource_lock_resource_type_query'
            ),
        }),
        'all_projects': parameter_types.multi_params({
            **parameter_types.boolean,
            'description': helpers.description(
                'resource_lock_all_projects_query'
            ),
        }),
        'lock_context': parameter_types.multi_params({
            'type': ['string', 'null'],
            'maxLength': 10,
            'description': helpers.description(
                'resource_lock_lock_context_query'
            ),
        }),
        'lock_reason': parameter_types.multi_params({
            'type': ['string', 'null'],
            'maxLength': 1023,
            'description': helpers.description(
                'resource_lock_lock_reason_query'
            ),
        }),
    },
    # TODO(stephenfin): Exclude additional query string parameters in a future
    # microversion
    'additionalProperties': True,
}

show_request_query = {
    'type': 'object',
    'properties': {},
    # TODO(stephenfin): Exclude additional query string parameters in a future
    # microversion
    'additionalProperties': True,
}

_resource_lock_response = {
    'type': 'object',
    'properties': {
        'id': {
            'type': 'string',
            'format': 'uuid',
            'description': helpers.description('resource_lock_id'),
        },
        'user_id': {
            'type': 'string',
            'format': 'uuid',
            'description': helpers.description('resource_lock_user_id'),
        },
        'project_id': {
            'type': 'string',
            'format': 'uuid',
            'description': helpers.description('resource_lock_project_id'),
        },
        'lock_context': {
            'type': 'string',
            'description': helpers.description('resource_lock_lock_context'),
        },
        'resource_type': {
            'type': 'string',
            'enum': constants.RESOURCE_LOCK_RESOURCE_TYPES,
            'description': helpers.description('resource_lock_resource_type'),
        },
        'resource_id': {
            'type': 'string',
            'format': 'uuid',
            'description': helpers.description('resource_lock_resource_id'),
        },
        'resource_action': {
            'type': 'string',
            'enum': constants.RESOURCE_LOCK_RESOURCE_ACTIONS,
            'description': helpers.description(
                'resource_lock_resource_action'
            ),
        },
        'lock_reason': {
            'type': ['string', 'null'],
            'description': helpers.description('resource_lock_lock_reason'),
        },
        'created_at': {
            'type': 'string',
            'format': 'date-time',
            'description': helpers.description('created_at'),
        },
        'updated_at': {
            'type': ['string', 'null'],
            'format': 'date-time',
            'description': helpers.description('updated_at'),
        },
        'links': response_types.links,
    },
    'description': helpers.description('resource_lock_object'),
    'required': [
        'id',
        'user_id',
        'project_id',
        'lock_context',
        'resource_type',
        'resource_id',
        'resource_action',
        'lock_reason',
        'created_at',
        'updated_at',
        'links',
    ],
    'additionalProperties': False,
}

create_response_body = {
    'type': 'object',
    'properties': {
        'resource_lock': _resource_lock_response,
    },
    'required': ['resource_lock'],
    'additionalProperties': False,
}

index_response_body = {
    'type': 'object',
    'properties': {
        'resource_locks': {
            'type': 'array',
            'items': _resource_lock_response,
        },
        'count': {
            'type': 'integer',
            'description': helpers.description('count_without_min_version'),
        },
        'resource_locks_links': response_types.collection_links,
    },
    'required': ['resource_locks'],
    'additionalProperties': False,
}

show_response_body = {
    'type': 'object',
    'properties': {
        'resource_lock': _resource_lock_response,
    },
    'required': ['resource_lock'],
    'additionalProperties': False,
}

update_response_body = {
    'type': 'object',
    'properties': {
        'resource_lock': _resource_lock_response,
    },
    'required': ['resource_lock'],
    'additionalProperties': False,
}

delete_response_body = {
    'type': 'null',
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The messages API controller module.

This module handles the following requests:
GET /messages
GET /messages/<message_id>
DELETE /messages/<message_id>
"""

from http import client as http_client

from oslo_utils import timeutils
import webob
from webob import exc

from manila.api import common
from manila.api.openstack import wsgi
from manila.api.schemas import messages as schema
from manila.api import validation
from manila.api.views import messages as messages_view
from manila import exception
from manila.i18n import _
from manila.message import api as message_api

MESSAGES_BASE_MICRO_VERSION = '2.37'
MESSAGES_QUERY_BY_TIMESTAMP = '2.52'


@validation.validated
class MessagesController(wsgi.Controller):
    """The User Messages API controller for the OpenStack API."""
    _view_builder_class = messages_view.ViewBuilder
    resource_name = 'message'

    def __init__(self):
        self.message_api = message_api.API()
        super(MessagesController, self).__init__()

    @wsgi.Controller.api_version(MESSAGES_BASE_MICRO_VERSION)
    @wsgi.Controller.authorize('get')
    @validation.request_query_schema(schema.show_request_query)
    @validation.response_body_schema(schema.show_response_body)
    def show(self, req, id):
        """Return the given message."""
        context = req.environ['manila.context']

        try:
            message = self.message_api.get(context, id)
        except exception.MessageNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        return self._view_builder.detail(req, message)

    @wsgi.Controller.api_version(MESSAGES_BASE_MICRO_VERSION)
    @wsgi.Controller.authorize
    @wsgi.action("delete")
    @validation.response_body_schema(schema.delete_response_body)
    def delete(self, req, id):
        """Delete a message."""
        context = req.environ['manila.context']

        try:
            message = self.message_api.get(context, id)
            self.message_api.delete(context, message)
        except exception.MessageNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        return webob.Response(status_int=http_client.NO_CONTENT)

    @wsgi.Controller.api_version(MESSAGES_BASE_MICRO_VERSION, '2.51')
    @wsgi.Controller.authorize('get_all')
    @validation.request_query_schema(schema.index_request_query)
    @validation.response_body_schema(schema.index_response_body)
    def index(self, req):
        """Returns a list of messages, transformed through view builder."""
        context = req.environ['manila.context']
        filters = req.params.copy()

        params = common.get_pagination_params(req)
        limit, offset = [params.get('limit'), params.get('offset')]
        sort_key, sort_dir = common.get_sort_params(filters)
        filters.pop('created_since', None)
        filters.pop('created_before', None)

        messages = self.message_api.get_all(context, search_opts=filters,
                                            limit=limit,
                                            offset=offset,
                                            sort_key=sort_key,
                                            sort_dir=sort_dir)

        return self._view_builder.index(req, messages)

    @wsgi.Controller.api_version(MESSAGES_QUERY_BY_TIMESTAMP)   # noqa: F811
    @wsgi.Controller.authorize('get_all')
    @validation.request_query_schema(schema.index_request_query_v252)
    @validation.response_body_schema(schema.index_response_body)
    def index(self, req):  # pylint: disable=function-redefined  # noqa F811
        """Returns a list of messages, transformed through view builder."""
        context = req.environ['manila.context']
        filters = req.params.copy()

        params = common.get_pagination_params(req)
        limit, offset = [params.get('limit'), params.get('offset')]
        sort_key, sort_dir = common.get_sort_params(filters)

        for time_comparison_filter in ['created_since', 'created_before']:
            if time_comparison_filter in filters:
                time_str = filters.get(time_comparison_filter)
                try:
                    parsed_time = timeutils.parse_isotime(time_str)
                except ValueError:
                    msg = _('Invalid value specified for the query '
                            'key: %s') % time_comparison_filter
                    raise exc.HTTPBadRequest(explanation=msg)

                filters[time_comparison_filter] = parsed_time

        messages = self.message_api.get_all(context, search_opts=filters,
                                            limit=limit,
                                            offset=offset,
                                            sort_key=sort_key,
                                            sort_dir=sort_dir)

        return self._view_builder.index(req, messages)


def create_resource():
    return wsgi.Resource(MessagesController())



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from manila.policies import base


BASE_POLICY_NAME = 'share_transfer:%s'

DEPRECATED_REASON = """
The transfer API now supports system scope and default roles.
"""

deprecated_share_transfer_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="Antelope"
)
deprecated_share_transfer_get_all_tenant = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all_tenant',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="Antelope"
)
deprecated_share_transfer_create = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="Antelope"
)
deprecated_share_transfer_get = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="Antelope"
)
deprecated_share_transfer_accept = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'accept',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="Antelope"
)
deprecated_share_transfer_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="Antelope"
)


share_transfer_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all',
        check_str=base.ADMIN_OR_PROJECT_READER,
        description="List share transfers.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-transfers'
            },
            {
                'method': 'GET',
                'path': '/share-transfers/detail'
            }
        ],
        deprecated_rule=deprecated_share_transfer_get_all
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all_tenant',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="List share transfers with all tenants.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-transfers'
            },
            {
                'method': 'GET',
                'path': '/share-transfers/detail'
            }
        ],
        deprecated_rule=deprecated_share_transfer_get_all_tenant
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        description="Create a share transfer.",
        operations=[
            {
                'method': 'POST',
                'path': '/share-transfers'
            }
        ],
        deprecated_rule=deprecated_share_transfer_create
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get',
        check_str=base.ADMIN_OR_PROJECT_READER,
        description="Show one specified share transfer.",
        operations=[
            {
                'method': 'GET',
                'path': '/share-transfers/{transfer_id}'
            }
        ],
        deprecated_rule=deprecated_share_transfer_get
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'accept',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        description="Accept a share transfer.",
        operations=[
            {
                'method': 'POST',
                'path': '/share-transfers/{transfer_id}/accept'
            }
        ],
        deprecated_rule=deprecated_share_transfer_accept
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        description="Delete share transfer.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/share-transfers/{transfer_id}'
            }
        ],
        deprecated_rule=deprecated_share_transfer_delete
    ),
]


def list_rules():
    return share_transfer_policies



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from webob import exc

from manila.db import base
from manila.share import rpcapi as share_rpcapi

CONF = cfg.CONF


class API(base.Base):
    """API for handling service actions."""

    def __init__(self):
        super(API, self).__init__()
        self.share_rpcapi = share_rpcapi.ShareAPI()

    def ensure_shares(self, context, service, host):
        """Start the ensure shares in a given host."""

        if service['state'] != "up":
            raise exc.HTTPConflict(
                "The service must have its state set to 'up' prior to running "
                "ensure shares.")

        self.share_rpcapi.ensure_driver_resources(context, host)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""backup_change_availability_zone_to_availability_zone_id

Revision ID: 2d708a9a3ba9
Revises: cb20f743ca7b
Create Date: 2023-08-24 11:01:41.134456

"""

# revision identifiers, used by Alembic.
revision = '2d708a9a3ba9'
down_revision = 'cb20f743ca7b'

from alembic import op
from sqlalchemy import Column, ForeignKey, String

from manila.db.migrations import utils


def collect_existing_az(az_table, connection):
    az_name_to_id_mapping = dict()
    for az in connection.execute(az_table.select()):
        if az.name in az_name_to_id_mapping:
            continue

        az_name_to_id_mapping[az.name] = az.id
    return az_name_to_id_mapping


def upgrade():
    connection = op.get_bind()

    op.add_column(
        'share_backups',
        Column('availability_zone_id', String(36),
               ForeignKey('availability_zones.id', name='sb_az_id_fk'))
    )

    # Collect existing AZs from availability_zones table
    availability_zones_table = utils.load_table(
        'availability_zones', connection)
    az_name_to_id_mapping = collect_existing_az(
        availability_zones_table, connection,)

    # Map string AZ names to ID's in target table
    # pylint: disable=no-value-for-parameter
    set_az_id_in_table = lambda table, id, name: (  # noqa: E731
        op.execute(
            table.update().where(table.c.availability_zone == name).values(
                {'availability_zone_id': id})
        )
    )

    share_backups_table = utils.load_table('share_backups', connection)
    for name, id in az_name_to_id_mapping.items():
        set_az_id_in_table(share_backups_table, id, name)

    # Remove old AZ columns from table
    op.drop_column('share_backups', 'availability_zone')


def downgrade():
    connection = op.get_bind()

    # Create old AZ fields
    op.add_column('share_backups',
                  Column('availability_zone', String(length=255)))

    # Migrate data
    az_table = utils.load_table('availability_zones', connection)
    share_backups_table = utils.load_table('share_backups', connection)

    for az in connection.execute(az_table.select()):
        # pylint: disable=no-value-for-parameter
        op.execute(
            share_backups_table.update().where(
                share_backups_table.c.availability_zone_id == az.id
            ).values({'availability_zone': az.name})
        )

    # Remove AZ_id columns and AZ table
    op.drop_constraint('sb_az_id_fk', 'share_backups', type_='foreignkey')
    op.drop_column('share_backups', 'availability_zone_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from manila import exception
from manila import ssh_utils
from manila import test
from oslo_utils import uuidutils
import paramiko
from unittest import mock


class FakeSock(object):
    def settimeout(self, timeout):
        pass


class FakeTransport(object):

    def __init__(self):
        self.active = True
        self.sock = FakeSock()

    def set_keepalive(self, timeout):
        pass

    def is_active(self):
        return self.active


class FakeSSHClient(object):

    def __init__(self):
        self.id = uuidutils.generate_uuid()
        self.transport = FakeTransport()

    def set_missing_host_key_policy(self, policy):
        pass

    def connect(self, ip, port=22, username=None, password=None,
                key_filename=None, look_for_keys=None, timeout=10,
                banner_timeout=10):
        pass

    def get_transport(self):
        return self.transport

    def close(self):
        pass

    def __call__(self, *args, **kwargs):
        pass


class SSHPoolTestCase(test.TestCase):
    """Unit test for SSH Connection Pool."""

    def test_single_ssh_connect(self):
        with mock.patch.object(paramiko, "SSHClient",
                               mock.Mock(return_value=FakeSSHClient())):
            sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10, "test",
                                        password="test", min_size=1,
                                        max_size=1)
            with sshpool.item() as ssh:
                first_id = ssh.id

            with sshpool.item() as ssh:
                second_id = ssh.id

            self.assertEqual(first_id, second_id)
            paramiko.SSHClient.assert_called_once_with()

    def test_create_ssh_with_password(self):
        fake_ssh_client = mock.Mock()
        ssh_pool = ssh_utils.SSHPool("127.0.0.1", 22, 10, "test",
                                     password="test")
        with mock.patch.object(paramiko, "SSHClient",
                               return_value=fake_ssh_client):
            ssh_pool.create()

            fake_ssh_client.connect.assert_called_once_with(
                "127.0.0.1", port=22, username="test",
                password="test", key_filename=None, look_for_keys=False,
                timeout=10, banner_timeout=10)

    def test_create_ssh_with_key(self):
        path_to_private_key = "/fakepath/to/privatekey"
        fake_ssh_client = mock.Mock()
        ssh_pool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                     "test",
                                     privatekey="/fakepath/to/privatekey")
        with mock.patch.object(paramiko, "SSHClient",
                               return_value=fake_ssh_client):
            ssh_pool.create()
            fake_ssh_client.connect.assert_called_once_with(
                "127.0.0.1", port=22, username="test", password=None,
                key_filename=path_to_private_key, look_for_keys=False,
                timeout=10, banner_timeout=10)

    def test_create_ssh_with_nothing(self):
        fake_ssh_client = mock.Mock()
        ssh_pool = ssh_utils.SSHPool("127.0.0.1", 22, 10, "test")
        with mock.patch.object(paramiko, "SSHClient",
                               return_value=fake_ssh_client):
            ssh_pool.create()
            fake_ssh_client.connect.assert_called_once_with(
                "127.0.0.1", port=22, username="test", password=None,
                key_filename=None, look_for_keys=True,
                timeout=10, banner_timeout=10)

    def test_create_ssh_error_connecting(self):
        attrs = {'connect.side_effect': paramiko.SSHException, }
        fake_ssh_client = mock.Mock(**attrs)
        ssh_pool = ssh_utils.SSHPool("127.0.0.1", 22, 10, "test")
        with mock.patch.object(paramiko, "SSHClient",
                               return_value=fake_ssh_client):
            self.assertRaises(exception.SSHException, ssh_pool.create)
            fake_ssh_client.connect.assert_called_once_with(
                "127.0.0.1", port=22, username="test", password=None,
                key_filename=None, look_for_keys=True,
                timeout=10, banner_timeout=10)

    def test_closed_reopend_ssh_connections(self):
        with mock.patch.object(paramiko, "SSHClient",
                               mock.Mock(return_value=FakeSSHClient())):
            sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                        "test", password="test",
                                        min_size=1, max_size=2)
            with sshpool.item() as ssh:
                first_id = ssh.id
            with sshpool.item() as ssh:
                second_id = ssh.id
                # Close the connection and test for a new connection
                ssh.get_transport().active = False
            self.assertEqual(first_id, second_id)
            paramiko.SSHClient.assert_called_once_with()

        # Expected new ssh pool
        with mock.patch.object(paramiko, "SSHClient",
                               mock.Mock(return_value=FakeSSHClient())):
            with sshpool.item() as ssh:
                third_id = ssh.id
            self.assertNotEqual(first_id, third_id)
            paramiko.SSHClient.assert_called_once_with()

    @mock.patch('builtins.open')
    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_sshpool_remove(self, mock_isfile, mock_sshclient, mock_open):
        ssh_to_remove = mock.Mock()
        mock_sshclient.side_effect = [mock.Mock(), ssh_to_remove, mock.Mock()]
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test", password="test",
                                    min_size=3, max_size=3)

        self.assertIn(ssh_to_remove, list(sshpool.free_items))

        sshpool.remove(ssh_to_remove)

        self.assertNotIn(ssh_to_remove, list(sshpool.free_items))

    @mock.patch('builtins.open')
    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_sshpool_remove_object_not_in_pool(self, mock_isfile,
                                               mock_sshclient, mock_open):
        # create an SSH Client that is not a part of sshpool.
        ssh_to_remove = mock.Mock()
        mock_sshclient.side_effect = [mock.Mock(), mock.Mock()]

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test", password="test",
                                    min_size=2, max_size=2)
        listBefore = list(sshpool.free_items)

        self.assertNotIn(ssh_to_remove, listBefore)

        sshpool.remove(ssh_to_remove)

        self.assertEqual(listBefore, list(sshpool.free_items))



# Copyright (c) 2022 MacroSAN Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

NFS_NON_CONFIG = 0
NFS_ENABLED = 1
NFS_DISABLED = 2
NFS_EXCEPTION = 3
NFS_NON_SUPPORTED = 0
NFS_SUPPORTED = 1

CIFS_SHARE_MODE = '2'
CIFS_ENABLED = '1'
CIFS_NON_CONFIG = '-1'
CIFS_DISABLED = '-2'
CIFS_EXCEPTION = '-3'

USER_NOT_EXIST = '0'
USER_EXIST = '1'
USER_FORMAT_ERROR = '2'

GROUP_NOT_EXIST = '0'
GROUP_EXIST = '1'
GROUP_FORMAT_ERROR = '2'

CODE_SUCCESS = 0
CODE_NOT_FOUND = 4
CODE_SOURCE_NOT_EXIST = 403

TOKEN_EXPIRED = 301
TOKEN_VERIFY_FAILED = 302
TOKEN_FORMAT_ERROR = 303
TOKEN_REQUIRED = 304
